/*
 * PAM module: Merge of pam_listfile with pam_mysql
 *
 * Rodrigo S Wanderley
 *
 * From MySQL module:
 *   Original Version written by: Gunay ARSLAN <arslan@gunes.medyatext.com.tr>
 *   This version by: James O'Kane <jo2y@midnightlinux.com>
 *   Modifications by Steve Brown, <steve@electronic.co.uk>
 *                  B.J. Black, <bj@taos.com>
 *                  Kyle Smith, <kyle@13th-floor.org>
 *                  Kees Cook, <kees@osdl.org>
 *
 * From listfile module:
 *   by Elliot Lee <sopwith@redhat.com>, Red Hat Software. July 25, 1996.
 *   log refused access error christopher mccrory <chrismcc@netus.com> 1998/7/11
 *
 *   This code began life as the pam_rootok module.
 */

/* turn on syslog debugging */
#undef DEBUG 
/* what level of syslog to use for debug reporting */
#define SYSLOG_DEBUG_LEVEL  LOG_ERR

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <stdarg.h>
#ifdef __linux__
#include <alloca.h>
#endif
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <mysql/mysql.h>

/*
 * here, we make definitions for the externally accessible functions
 * in this file (these definitions are required for static modules
 * but strongly encouraged generally) they are used to instruct the
 * modules include file to define their prototypes.
 */

const char * pam_module_name = "pam_listmysql";

#define PAM_SM_AUTH

#define PAM_MODULE_NAME             (pam_module_name)

#ifdef DEBUG
# define DEBUG_syslog(t, s...)  syslog(SYSLOG_DEBUG_LEVEL, t, ##s)
#else
# define DEBUG_syslog(t, s...)  /* */
#endif

/* Extended Items that are not directly available via pam_get_item() */
#define EI_GROUP (1 << 0)
#define EI_SHELL (1 << 1)

#include <security/pam_modules.h>
#include <security/pam_misc.h>

typedef struct pam_listmysql_data_t {
  MYSQL *mysql_auth;		/* these are for mysql */
  char  *host;
  char  *where;
  char  *database;
  char  *dbuser;
  char  *dbpasswd;
  char  *table;
  char  *column;

  int   citem;			/* type of item */
  int   extitem;		/* item not available directly by pam */
  int   sense;			/* allow|deny */
  char  *itemlist[256];		/* used for item = group */
} pam_listmysql_data;

/* 
 * General PAM Prototypes 
 */
int converse (pam_handle_t * pamh,
	      int nargs,
	      struct pam_message **message,
	      struct pam_response **response);

int _set_auth_tok (pam_handle_t * pamh,
		   int flags, int argc,
		   const char **argv);

PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh,
				   int flags, int argc, const char **argv);
PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh, int flags, int argc,
			      const char **argv);

/*
 * Internal function Prototypes
 */
pam_listmysql_data * parseArgs(pam_handle_t *pamh, int argc, const char **argv);
void pam_listmysql_cleanup(pam_handle_t *pamh, void * data, int error_status);

/*
 * MySQL Related Prototypes
 */
int db_connect(pam_listmysql_data * data);
void db_close (pam_listmysql_data * data);

pam_listmysql_data *
parseArgs(pam_handle_t *pamh, int argc, const char **argv) 
{
  pam_listmysql_data * data=NULL;
  char *option=NULL;
  char *value=NULL;
  int  i, size;

  pam_get_data(pamh,PAM_MODULE_NAME,(const void**)&data);
  if (!data) {
    /* allocate global data space */
    data=(pam_listmysql_data*)calloc(1,sizeof(pam_listmysql_data));
    if (!data) {
      syslog(LOG_CRIT,"Out of memory in parseArgs");
      return NULL;
    }

    /* give the data back to PAM for management */
    if (pam_set_data(pamh,PAM_MODULE_NAME,
		     (void*)data,pam_listmysql_cleanup)!=PAM_SUCCESS) {
      syslog(LOG_CRIT,"pam did not take our module data");
      free(data);
      return NULL;
    }
  }
  /* close the database in case we get new args */
  db_close(data);

  /* set up the default arguments */
  data->host         =strdup("localhost");
  data->where        =strdup("");
  data->database     =strdup("mysql");
  data->dbuser       =strdup("nobody");
  data->dbpasswd     =strdup("");
  data->table        =strdup("user");
  data->column       =strdup("user.User");
  data->citem        =PAM_USER;
  data->extitem      =0;
  data->sense        =0;
  data->itemlist[0]  =0;

  /* process all the arguments */
  for (i=0; i<argc; i++) {
    /* since argv is technically a const char, we shouldn't
     * assume that we can write in it.  To be safe, we'll
     * keep our own copies of everything instead of using strtok.
     */
    if ((value=index(argv[i],'='))!=NULL) {
      size=value-argv[i]; /* length of the option name */
            
      if (option) free(option); /* unalloc previous option name */
      option=(char*)malloc(size+1); /* to store option name + NULL */
      strncpy(option,argv[i],size);
      option[size]='\0'; /* make sure we're NULL terminated */

      value++; /* get past the '=' */

      /* set the various options */
      if (!strcasecmp ("host", option)) {
	if (data->host) free(data->host);
	data->host=strdup(value);
	DEBUG_syslog("host changed.");
      } else if (!strcasecmp ("where", option)) {
	if (data->where) free(data->where);
	data->where=strdup(value);
	DEBUG_syslog("where changed.");
      } else if (!strcasecmp ("db", option)) {
	if (data->database) free(data->database);
	data->database=strdup(value);
	DEBUG_syslog("database changed.");
      } else if (!strcasecmp ("user", option)) {
	if (data->dbuser) free(data->dbuser);
	data->dbuser=strdup(value);
	DEBUG_syslog("dbuser changed.");
      } else if (!strcasecmp ("passwd", option)) {
	if (data->dbpasswd) free(data->dbpasswd);
	data->dbpasswd=strdup(value);
	DEBUG_syslog("dbpasswd changed.");
      } else if (!strcasecmp ("table", option)) {
	if (data->table) free(data->table);
	data->table=strdup(value);
	DEBUG_syslog("table changed.");
      } else if (!strcasecmp ("column", option)) {
	if (data->column) free(data->column);
	data->column=strdup(value);
	DEBUG_syslog("column changed.");
      } else if(!strcasecmp("item", option)) {
	if(!strcasecmp("user", value)) {
	  data->citem = PAM_USER;
	} else if(!strcasecmp("tty", value)) {
	  data->citem = PAM_TTY;
	} else if(!strcmp("rhost", value)) {
	  data->citem = PAM_RHOST;
	} else if(!strcmp("ruser", value)) {
	  data->citem = PAM_RUSER;
	} else {		/* these items are related to the user,
				   but are not directly gettable with
				   pam_get_item */
	  data->citem = PAM_USER;
	  if (!strcasecmp("group", value)) {
	    data->extitem = EI_GROUP;
	  } else if(!strcmp("shell", value)) {
	    data->extitem = EI_SHELL;
	  } else {
	    syslog(LOG_ERR, "Unknown item: %s, using user!", value);
	  }
	}
	DEBUG_syslog("type changed: %d (%s)", data->citem, value);
      } else if(!strcasecmp("sense", option)) {
	if(!strcasecmp(value, "allow")) {
	  data->sense = 0;
	} else if(!strcasecmp(value, "deny")) {
	  data->sense = 1;
	}
	DEBUG_syslog("sense changed: %d", data->sense);
      } else {
	syslog(LOG_ERR, "Unknown option: %s=%s", option, value);
      }
    } else {
      syslog(LOG_ERR, "Unknown option: %s", argv[i]);
    }
  } /* for loop */
  if (option) free(option); /* clean up option name allocation */

  return data;
}


/* MySQL access functions */
int 
db_connect (pam_listmysql_data * data) 
{
  int retvalue = PAM_AUTH_ERR;

  DEBUG_syslog("db_connect  called.");

  /* valid data? */
  if (!data)
    return PAM_AUTH_ERR;

  /* already connected? */
  if (data->mysql_auth != NULL)
    return PAM_SUCCESS;

  if (!(data->mysql_auth=mysql_init(NULL))) {
    syslog(LOG_CRIT,"Out of memory!");
    return PAM_AUTH_ERR;
  }
  
  if (mysql_real_connect(data->mysql_auth, data->host,
			 data->dbuser, data->dbpasswd,
			 data->database, 0, NULL, 0)) {
    if (!mysql_select_db(data->mysql_auth, data->database))
      retvalue = PAM_SUCCESS;
  }
  
  if (retvalue != PAM_SUCCESS)
    syslog(LOG_ERR, "pam_listmysql: MySQL err %s", mysql_error(data->mysql_auth));

  DEBUG_syslog("returning %i .",retvalue);

  return retvalue;
}

void 
db_close(pam_listmysql_data * data) 
{
  if (!data || !data->mysql_auth)
    return; /* closed already */

  /* this unallocates for us, since we used mysql_init(NULL) */
  mysql_close(data->mysql_auth);
  data->mysql_auth = NULL;
}

static int 
db_checklist(pam_listmysql_data * data, const char *item)
{
  char *sql = NULL;		/* Buffer for SQL query */
  int querysize = 0;		/* Calculated size of query */
  int nrows = 0;		/* # of rows */
  int i;
  MYSQL_RES *result;
  MYSQL_ROW row;
  int retvalue = PAM_AUTH_ERR;

  DEBUG_syslog("db_checklist called.");

  if (!data) {
    syslog(LOG_ERR,"no global data");
    return PAM_AUTH_ERR;
  }

  querysize = strlen("select  from  where  ") + 
    strlen(data->column) +
    strlen(data->table) +
    strlen(data->where);

  sql = malloc(sizeof(char) * querysize);
  if (sql == NULL)
    return PAM_BUF_ERR;

  snprintf(sql, querysize, "SELECT %s FROM %s",
	   data->column, data->table);

  if (strlen(data->where) > 0) {
    strncat(sql, " WHERE ", (querysize - strlen(sql)));
    strncat(sql, data->where, (querysize - strlen(sql)));
  }

  DEBUG_syslog("pam_listmysql: where clause = %s", data->where);

  /* These lines are commented even though its inside a ifdef DEBUG
     Why? Because it exposes user names and passwords into syslog and is
     a MASSIVE security hole. Uncomment only if you REALLY need to see
     the query. Remember, usernames and attmpted passwords will be
     exposed into syslog though!
  */

  DEBUG_syslog("pam_listmysql: %s", sql);

  mysql_query(data->mysql_auth, sql);
  free(sql);
  sql=NULL;
  
  result = mysql_store_result(data->mysql_auth);
  if (result == NULL) {
    syslog(LOG_ERR, "%s", mysql_error(data->mysql_auth));
    return PAM_AUTH_ERR;
  }

  nrows = mysql_num_rows(result);
  DEBUG_syslog("got %d rows.", nrows);

  while (nrows--) {
    row = mysql_fetch_row(result);
    
    if (row == NULL) {
      syslog(LOG_ERR, "%s", mysql_error(data->mysql_auth));
      mysql_free_result(result);
      return PAM_AUTH_ERR;
    }

    if (row[0] == NULL) {
      DEBUG_syslog("returning %i .", PAM_AUTH_ERR);
      mysql_free_result(result);
      return PAM_AUTH_ERR;
    }

    if (data->extitem == EI_GROUP) {
      for (i = 0; data->itemlist[i]; ++i) {
	DEBUG_syslog("row[0]: %s itemlist: %s", row[0], data->itemlist[i]);
	if (!strcmp(data->itemlist[i], row[0])) {
	  retvalue = PAM_SUCCESS;
	  goto group_matched;
	}
      } 
    } else {
      DEBUG_syslog("row[0]: %s item: %s", row[0], item);
      if (!strcmp(item, row[0])) {
	retvalue = PAM_SUCCESS;
	break;
      }
    }
  }

 group_matched:
  DEBUG_syslog("returning %i .",retvalue);
  mysql_free_result(result);
  return retvalue;
}


/* Global PAM functions stolen from other modules */

int 
converse(pam_handle_t *pamh, int nargs, struct pam_message **message,
	 struct pam_response **response)
{
  int retval;
  struct pam_conv *conv;

  retval = pam_get_item(pamh, PAM_CONV, (const void **) &conv); 

  if (retval == PAM_SUCCESS) {
    retval = conv->conv(nargs,
			(const struct pam_message **) message,
			response, conv->appdata_ptr);
    if ((retval != PAM_SUCCESS) && (retval != PAM_CONV_AGAIN))
      syslog(LOG_ERR, "pam_listmysql: conversation failure [%s]",
	     pam_strerror(pamh, retval));
  } else {
    syslog(LOG_ERR, "pam_listmysql: couldn't obtain coversation function [%s]",
	   pam_strerror(pamh, retval));
  }

  return retval;	/* propagate error status */
}

/* checks if a user is on a list of members */
static int 
is_on_list(char * const *list, const char *member)
{
    while (*list) {
        if (strcmp(*list, member) == 0)
            return 1;
        list++;
    }
    return 0;
}

/* PAM Authentication functions */

PAM_EXTERN int 
pam_sm_authenticate (pam_handle_t * pamh,
		     int flags,
		     int argc,
		     const char **argv)
{
  pam_listmysql_data * data;
  int retval, i;
  const char *citemp;
#ifdef HORRIBLE_HACK_FOR_MYSQL_3_23_27
  char mem[1000];
#endif

  /* Stuff for "extended" items */
  struct passwd *userinfo;
  struct group *grpinfo;

  DEBUG_syslog("pam_sm_authenticate called.");

  if (!(data=parseArgs(pamh, argc, argv))) {
      syslog(LOG_ERR,"failed to load options");
      return PAM_AUTH_ERR;
  }

  /* Get Item */

  retval = pam_get_item(pamh, data->citem, (const void **)&citemp);
  if (retval != PAM_SUCCESS) {
    return retval;
  }
  if ((data->citem == PAM_USER) && !citemp) {
    pam_get_user(pamh, &citemp, NULL);
    if (retval != PAM_SUCCESS)
      return PAM_SERVICE_ERR;
  }

  if (retval != PAM_SUCCESS || citemp == NULL) {
    syslog(LOG_ERR, "%s", "pam_listmysql: no user specified");
    DEBUG_syslog("returning.");
    return PAM_USER_UNKNOWN;
  }

  if (strlen(citemp) <= 0)	/* the item is NULL, no match */
    return data->sense?PAM_SUCCESS:PAM_AUTH_ERR;

  if (data->extitem) {
    switch (data->extitem) {
    case EI_GROUP:
      userinfo = getpwnam(citemp);
      if (userinfo == NULL) {
	syslog(LOG_ERR, "getpwnam(%s) failed", citemp);
	return PAM_SERVICE_ERR;
      }
      grpinfo = getgrgid(userinfo->pw_gid);
      if (grpinfo == NULL) {
	syslog(LOG_ERR, "getgrpid(%d) failed",
	       (int)userinfo->pw_gid);
	return PAM_SERVICE_ERR;
      }
      DEBUG_syslog("adding group %s to itemlist[0]", grpinfo->gr_name);
      data->itemlist[0] = strdup(grpinfo->gr_name);
      if (!data->itemlist[0])
	return PAM_SERVICE_ERR;

      setgrent();
      for (i = 1; (i < sizeof(data->itemlist)/sizeof(data->itemlist[0])-1) &&
	     (grpinfo = getgrent()); ) {
	if (is_on_list(grpinfo->gr_mem, citemp)) {
	  DEBUG_syslog("adding group %s to itemlist[%d]", grpinfo->gr_name, i);
	  data->itemlist[i] = strdup(grpinfo->gr_name);
	  if (!data->itemlist[i++])
	    return PAM_SYSTEM_ERR;
	}
      }
      endgrent();
      DEBUG_syslog("%d groups added", i);
      data->itemlist[i] = NULL;
      break;
    case EI_SHELL:
      userinfo = getpwnam(citemp);
      if (userinfo == NULL) {
	syslog(LOG_ERR, "getpwnam(%s) failed", citemp);
	return PAM_SYSTEM_ERR;
      }
      citemp = userinfo->pw_shell;
      break;
    default:
      syslog(LOG_ERR, "Internal weirdness, unknown extended item %d",
	     data->extitem);
      return PAM_SYSTEM_ERR;
    }
  }
	
  if ((retval = db_connect(data)) != PAM_SUCCESS) {
    DEBUG_syslog("returning %i after db_connect.",retval);
    return retval;
  }

  retval = db_checklist(data, citemp);
  if (retval != PAM_SUCCESS) {
    return data->sense ? PAM_SUCCESS : PAM_AUTH_ERR;
  } else {
    return data->sense ? PAM_AUTH_ERR : PAM_SUCCESS;
  }
} /* pam_sm_authenticate */


PAM_EXTERN int 
pam_sm_setcred(pam_handle_t *pamh,int flags,int argc
	       ,const char **argv)
{
  DEBUG_syslog("pam_listmysql: setcred called but not implemented.");
  return PAM_SUCCESS;
}

/* unallocates any used memory */
void 
pam_listmysql_cleanup(pam_handle_t *pamh, void * voiddata, int error_status)
{
  pam_listmysql_data * data;
  data=(pam_listmysql_data*)voiddata;
  int i;

  if (data) {
    db_close(data);
    if (data->host) free(data->host);
    if (data->where) free(data->where);
    if (data->database) free(data->database);
    if (data->dbuser) free(data->dbuser);
    if (data->dbpasswd) free(data->dbpasswd);
    if (data->table) free(data->table);
    if (data->column) free(data->column);

    for (i = 0; data->itemlist[i]; ++i)
      free(data->itemlist[i]);

    free(data);
  }
}

/* end of module definition */

#ifdef PAM_STATIC

/* static module data */

struct pam_module _pam_listmysql_modstruct = {
  "pam_listmysql",
  pam_sm_authenticate,
  pam_sm_setcred,
  NULL,
  NULL,
  NULL,
  NULL
};

#endif
